import pytest

from flask import Flask, request

from odinweb import flask
from odinweb.constants import Method, Type
from odinweb.data_structures import PathParam


def test_request_proxy():
    test_app = Flask(__name__)
    test_app.testing = True
    app = test_app.test_client()

    @test_app.route('/')
    def test_method():
        target = flask.RequestProxy(request)
        assert target.method == Method.POST
        assert set(target.query.getlist('a')) == {'1', '3'}
        assert set(target.query.getlist('b')) == {'2'}
        assert target.body == "123"
        assert target.content_type == 'text/html'
        assert target.origin == "http://localhost"
        return 'OK'

    app.post('/?a=1&b=2&a=3', content_type='text/html', data="123", headers={"Origin": 'http://localhost'})


class TestApiBlueprint(object):
    @pytest.mark.parametrize('node, expected', (
        (PathParam('foo'), '<int:foo>'),
        (PathParam('bar', Type.String), '<string:bar>'),
        (PathParam('bar', None, None), '<bar>'),
    ))
    def test_node_formatter(self, node, expected):
        actual = flask.ApiBlueprint.node_formatter(node)
        assert actual == expected



import odin

from flask import Flask
from odin.utils import getmeta, field_iter_items, field_iter

from odinweb import api, doc
from odinweb.data_structures import HttpResponse
from odinweb.flask import ApiBlueprint
from odinweb.cors import CORS
from odinweb.swagger import SwaggerSpec


class User(odin.Resource):
    """
    User resource
    """
    id = odin.IntegerField(key=True)
    username = odin.StringField()
    name = odin.StringField()
    email = odin.EmailField()
    role = odin.StringField(choices=('a', 'b', 'c'))


class UserSummary(odin.ResourceProxy):
    class Meta:
        resource = User
        include = ('id', 'username', 'name')
        readonly = ('id', 'username')


USERS = [
    User(1, 'pimpstar24', 'Bender', 'Rodreges', 'bender@ilovebender.com'),
    User(2, 'zoidberg', 'Zoidberg', '', 'zoidberg@freemail.web'),
    User(3, 'amylove79', 'Amy', 'Wong', 'awong79@marslink.web'),
]
USER_ID = len(USERS)


class UserApi(api.ResourceApi):
    resource = User
    tags = ['user']

    @doc.deprecated
    @api.collection(path='find', methods=api.Method.POST)
    def operation_test(self, request):
        pass

    @api.listing(resource=UserSummary)
    def get_user_list(self, request, offset, limit):
        return UserSummary.proxy(USERS[offset:offset+limit]), len(USERS)

    @api.action(path='stream')
    def get_user_stream(self, request):
        """
        Streamed user response.
        """
        def generate():
            yield '\t'.join(f.name for f in field_iter(User)) + '\n'

            for user in USERS:
                yield '\t'.join(str(v) for _, v in field_iter_items(user)) + '\n'

        return HttpResponse(generate(), headers={'Content-Type': 'text/plain'})

    @api.create(resource=UserSummary)
    def create_user(self, request, user):
        global USER_ID

        # Add user to list
        USER_ID += 1
        user.id = USER_ID
        USERS.append(user)

        return user

    @api.detail
    def get_user(self, request, id):
        """
        Get a user object
        """
        for user in USERS:
            if user.id == id:
                return user

        raise api.HttpError(api.HTTPStatus.NOT_FOUND)

    @api.update
    def update_user(self, request, user, id):
        return user

    @api.patch
    def patch_user(self, request, user, id):
        return user

    @api.delete
    def delete_user(self, request, id):
        for idx, user in enumerate(USERS):
            if user.id == id:
                USERS.remove(user)
                return

        raise api.HttpError(api.HTTPStatus.NOT_FOUND)


sample_api = api.ApiCollection(name='sample')


@sample_api.operation(path='foo/bar')
def sample(request):
    return {}


app = Flask(__name__)
app.register_blueprint(
    CORS(
        ApiBlueprint(
            api.ApiVersion(
                SwaggerSpec("Flask Example Swaggerspec", enable_ui=True),
                sample_api,
                UserApi(),
            ),
            middleware=[],
            debug_enabled=True,
        ),
        origins=['http://localhost:5000'],
        max_age=10,
    )
)


if __name__ == '__main__':
    app.run(debug=True)



"""
OdinWeb.Flask API
~~~~~~~~~~~~~~~~~

Flask implementation of the OdinWeb API interface.

The API integrates into Flask via Flasks blueprint features.

"""
from __future__ import absolute_import

from flask import request, session, Response

from odin.utils import lazy_property
from odinweb.containers import ApiInterfaceBase
from odinweb.constants import Type, Method
from odinweb.data_structures import MultiValueDict, BaseHttpRequest

# Type imports
from flask import Flask, Request  # noqa
from odinweb.data_structures import PathParam  # noqa


TYPE_MAP = {
    Type.Integer: 'int',
    Type.Long: 'int',
    Type.Float: 'number',
    Type.Double: 'number',
    Type.String: 'string',
    Type.Byte: '',
    Type.Binary: '',
    Type.Boolean: 'bool',
    Type.Date: 'string',
    Type.Time: 'string',
    Type.DateTime: 'string',
    Type.Password: 'string',
}


class RequestProxy(BaseHttpRequest):
    def __init__(self, r):
        # type: (Request) -> None
        self.request = r

    @lazy_property
    def environ(self):
        return self.request.environ

    @lazy_property
    def method(self):
        try:
            return Method(self.request.method)
        except KeyError:
            pass

    @lazy_property
    def scheme(self):
        return self.request.scheme

    @lazy_property
    def host(self):
        return self.request.host

    @lazy_property
    def path(self):
        return self.request.path

    @lazy_property
    def query(self):
        return MultiValueDict(self.request.args)

    @lazy_property
    def headers(self):
        return self.request.headers

    @lazy_property
    def cookies(self):
        return self.request.cookies

    @lazy_property
    def session(self):
        return session

    @lazy_property
    def body(self):
        return self.request.data

    @lazy_property
    def form(self):
        return MultiValueDict(self.request.form)


class ApiBlueprint(ApiInterfaceBase):
    """
    A Flask Blueprint for an API::

        from flask import Flask
        from odinweb.flask.api import ApiBlueprint

        app = Flask(__name__)

        app.register_blueprint(
            ApiBlueprint(
                ApiVersion(
                    UserApi(),
                    version='v1
                )
            )
        )

    """
    def __init__(self, *containers, **options):
        self.subdomain = options.pop('subdomain', None)
        super(ApiBlueprint, self).__init__(*containers, **options)

    @staticmethod
    def node_formatter(path_node):
        # type: (PathParam) -> str
        """
        Format a node to be consumable by the `UrlPath.parse`.
        """
        if path_node.type:
            node_type = TYPE_MAP.get(path_node.type, 'str')
            return "<{}:{}>".format(node_type, path_node.name)
        return "<{}>".format(path_node.name)

    def _bound_callback(self, operation):
        def callback(**path_args):
            response = self.dispatch(operation, RequestProxy(request), **path_args)
            return Response(response.body or ' ', response.status, response.headers)
        callback.provide_automatic_options = False
        return callback

    def register(self, app, options, first_registration):
        # type: (Flask, dict, bool) -> None
        """
        Register interface

        :param app: Instance of flask.
        :param options: Options for blueprint
        :param first_registration: First registration of blueprint

        """
        for url_path, operation in self.op_paths():
            app.add_url_rule(
                url_path.format(self.node_formatter),
                '%s.%s' % (self.name, operation.operation_id),
                self._bound_callback(operation),
                methods=(m.value for m in operation.methods),
                **options
            )
